/**
 * Instrumentation Vuze Plugin
 * 
 * Copyright 2008, Yale Laboratory of Networked Systems
 */

package net.yale.cs.p4p.p2p.azureus.log;

import java.util.Collections;
import java.util.Map;
import java.util.TreeMap;

import net.yale.cs.p4p.p2p.azureus.log.listeners.XFerDownloadListener;
import net.yale.cs.p4p.p2p.azureus.log.listeners.XFerDownloadPeerListener;
import net.yale.cs.p4p.p2p.azureus.log.records.DownloadLogPoster;
import net.yale.cs.p4p.p2p.logs.DownloadLogs.ComponentIdentity;
import net.yale.cs.p4p.p2p.logs.DownloadLogs.ComponentParams;
import net.yale.cs.p4p.p2p.logs.DownloadLogs.Components;

import org.gudy.azureus2.plugins.PluginEvent;
import org.gudy.azureus2.plugins.PluginEventListener;
import org.gudy.azureus2.plugins.PluginException;
import org.gudy.azureus2.plugins.PluginInterface;
import org.gudy.azureus2.plugins.PluginListener;
import org.gudy.azureus2.plugins.UnloadablePlugin;
import org.gudy.azureus2.plugins.ipc.IPCException;
import org.gudy.azureus2.plugins.ipc.IPCInterface;
import org.gudy.azureus2.plugins.logging.LoggerChannel;

public class XFerLogPlugin implements UnloadablePlugin
{
	public static final String PLUGIN_NAME = "Instrumentation";

	public static final String IPC_METHOD_GET_PARAMS = "getInstrumentationParams";

	
	private class XFerLogPluginEventListener implements PluginEventListener {
		public void handleEvent(PluginEvent ev) {
			switch (ev.getType()) {
			case PluginEvent.PEV_ALL_PLUGINS_INITIALISED:
				updateComponents();
				break;
			}
		}
	}

	private class XFerLogPluginListener implements PluginListener {
		public void closedownComplete() {
			finishUnload();
		}

		public void closedownInitiated() {
			startUnload();
			finishUnload();
		}
		public void initializationComplete() { }
	}
	
	/* Configuration options */
	private Config config_;
	public Config getConfig() { return config_; }

	/* Current set of registered algorithms */
	private Components components_ = new Components();
	public Components getComponents() { return components_; }

	/* Azureus plugin interface */
	private PluginInterface pluginInterface_;
	public PluginInterface getPluginInterface() { return pluginInterface_; }

	/* Our very own logger channel */
	private LoggerChannel logger_;
	public LoggerChannel getLogger() { return logger_; }
	
	/* Log record processor for posting records to logging server */
	private DownloadLogPoster processChain_;
	public DownloadLogPoster getLogProcessChain() { return processChain_; }

	/* Event listeners */
	private XFerLogPluginListener pluginListener_;
	private XFerLogPluginEventListener pluginEventListener_;
	private XFerDownloadListener downloadListener_;
	private XFerDownloadPeerListener downloadPeerListener_;

	public void initialize(PluginInterface pluginInterface) throws PluginException {
		pluginInterface_ = pluginInterface;

		/* Create a logger for our plugin */
		logger_ = pluginInterface_.getLogger().getChannel(PLUGIN_NAME);
		logger_.log("initializing plugin");
		
		/* Create configuration options */
		logger_.log("installing configuration");
		config_ = new Config(this);

		/* Create chain for processing logs */
		processChain_ = new DownloadLogPoster(this, null);

		/* Create and install listeners */
		logger_.log("creating listeners");
		pluginListener_ = new XFerLogPluginListener();
		pluginEventListener_ = new XFerLogPluginEventListener();
		downloadListener_ = new XFerDownloadListener(this);
		downloadPeerListener_ = new XFerDownloadPeerListener(this);

		logger_.log("installing listeners");
		pluginInterface_.addListener(pluginListener_);
		pluginInterface_.addEventListener(pluginEventListener_);
		pluginInterface_.getDownloadManager().getGlobalDownloadEventNotifier().addPeerListener(downloadPeerListener_);
		pluginInterface_.getDownloadManager().getGlobalDownloadEventNotifier().addListener(downloadListener_);
		
		logger_.log("looking for components");
		updateComponents();

		logger_.log("finished initialization");
	}

	private void startUnload() {
		/* Remove listeners */
		if (logger_ != null)
			logger_.log("removing listeners");
		if (downloadListener_ != null) {
			downloadListener_.cleanup();
			pluginInterface_.getDownloadManager().getGlobalDownloadEventNotifier().removeListener(downloadListener_);
			downloadListener_ = null;
		}
		if (downloadPeerListener_ != null) {
			downloadPeerListener_.cleanup();
			pluginInterface_.getDownloadManager().getGlobalDownloadEventNotifier().removePeerListener(downloadPeerListener_);
			downloadPeerListener_ = null;
		}
		if (pluginEventListener_ != null) {
			pluginInterface_.removeEventListener(pluginEventListener_);
			pluginEventListener_ = null;
		}
		if (pluginListener_ != null) {
			pluginInterface_.removeListener(pluginListener_);
			pluginListener_ = null;
		}

		/* Stop threads */
		if (logger_ != null)
			logger_.log("stopping threads");
		if (processChain_ != null)
			processChain_.stop();
	}

	private void finishUnload() {
		/* Wait for threads */
		if (logger_ != null)
			logger_.log("waiting for threads");
		if (processChain_ != null) {
			processChain_.join();
			processChain_ = null;
		}

		/* Clean up configuration */
		if (logger_ != null)
			logger_.log("removing configuration");
		if (config_ != null) {
			config_.cleanup();
			config_ = null;
		}

		if (logger_ != null) {
			logger_.log("unloaded plugin");
			logger_ = null;
		}
	}

	public void unload() throws PluginException {
		startUnload();
		finishUnload();
	}

	private String getStringToken(Object o) {
		if (o == null)
			return "null";
		String s = o.toString();
		if (s.length() == 0)
			return "null";
		return s.replaceAll("\\s+", "_");
	}

	private Map<String, String> getPluginParams(PluginInterface plugin) {
		/* Get IPC interface to the plugin */
		IPCInterface ipc = plugin.getIPC();
		if (ipc == null)
			return Collections.emptyMap();

		/* Query the plugin for any parameters it gives */
		Object pluginResult;
		try {
			pluginResult = ipc.invoke(IPC_METHOD_GET_PARAMS, new Object[] {});
		} catch (IPCException e) {
			return Collections.emptyMap();
		}

		/* Ensure result is of expected type */
		if (!(pluginResult instanceof Map))
			return Collections.emptyMap();

		/* Convert into a map of the correct type */
		Map<String, String> result = new TreeMap<String, String>();
		for (Object eObj : ((Map)pluginResult).entrySet()) {
			Map.Entry e = (Map.Entry)eObj;
			result.put(
				getStringToken(e.getKey()),
				getStringToken(e.getValue()));
		}
		return result;
	}

	private Object updateComponentsLock_ = new Object();
	
	public void updateComponents() {
		synchronized (updateComponentsLock_) {
			Components result = new Components();
			for (PluginInterface plugin : getPluginInterface().getPluginManager().getPlugins())
				result.put(
					new ComponentIdentity(getStringToken(plugin.getPluginID()), getStringToken(plugin.getPluginVersion())),
					new ComponentParams(getPluginParams(plugin)));
			components_ = result;

			getLogProcessChain().resetNeedConfig();
		}
	}

}
