/**
 * Instrumentation Vuze Plugin
 * 
 * Copyright 2008, Yale Laboratory of Networked Systems
 */

package net.yale.cs.p4p.p2p.azureus.log.listeners;

import net.yale.cs.p4p.p2p.azureus.log.XFerLogPlugin;
import net.yale.cs.p4p.p2p.azureus.log.records.PeerTransferStats;
import net.yale.cs.p4p.p2p.azureus.log.records.PluginDownloadLogs;
import net.yale.cs.p4p.p2p.logs.DownloadLogs;
import net.yale.cs.p4p.p2p.logs.DownloadLogs.Event;

import org.gudy.azureus2.plugins.download.Download;
import org.gudy.azureus2.plugins.messaging.Message;
import org.gudy.azureus2.plugins.network.Connection;
import org.gudy.azureus2.plugins.network.IncomingMessageQueue;
import org.gudy.azureus2.plugins.network.IncomingMessageQueueListener;
import org.gudy.azureus2.plugins.network.OutgoingMessageQueue;
import org.gudy.azureus2.plugins.network.OutgoingMessageQueueListener;
import org.gudy.azureus2.plugins.peers.Peer;
import org.gudy.azureus2.plugins.peers.PeerEvent;
import org.gudy.azureus2.plugins.peers.PeerListener2;

public class XFerPeerListener implements PeerListener2 {

	/* Detect activity for incoming (downloaded) content */
	private class XFerIncomingMessageQueueListener implements IncomingMessageQueueListener {
		public void bytesReceived(int byte_count) {
			if (peerDownloadStats_ == null)
				peerDownloadStats_ = new PeerTransferStats(peer_);
			else
				handleNewMessage(Event.EVT_XFER_DLOAD_STATS, peerDownloadStats_, false);
		}
		public boolean messageReceived(Message message) { return false; }
	}

	/* Detect activity for outgoing (uploaded) content */
	private class XFerOutgoingMessageQueueListener implements OutgoingMessageQueueListener {
		public void bytesSent(int byte_count) {
			if (peerUploadStats_ == null)
				peerUploadStats_ = new PeerTransferStats(peer_);
			else
				handleNewMessage(Event.EVT_XFER_UPLOAD_STATS, peerUploadStats_, false);
		}
		public boolean messageAdded(Message message) { return true; }
		public void messageSent(Message message) { }
	}

	private void handleNewMessage(Event event, PeerTransferStats stats, boolean forceSend) {
		if (forceSend || System.currentTimeMillis() - stats.getEndTime() >= THRESHOLD_TIME) {
			/* We had a previous record that was out of date */
			if (Event.EVT_XFER_DLOAD_STATS.equals(event) && stats.hasDownloadData()
					|| Event.EVT_XFER_UPLOAD_STATS.equals(event) && stats.hasUploadData())
				plugin_.getLogProcessChain().push(
						PluginDownloadLogs.MakeLogRecordTransferEvent(plugin_, event, peer_, stats));
			stats.next();
			stats.update(peer_);
		} else {
			/* We had a previous record but it was still within the threshold */
			stats.update(peer_);
		}
	}

	private long THRESHOLD_TIME = 10 * 1000;

	private XFerLogPlugin plugin_;
	private Peer peer_;

	private IncomingMessageQueueListener incomingMsgListener_ = new XFerIncomingMessageQueueListener();
	private OutgoingMessageQueueListener outgoingMsgListener_ = new XFerOutgoingMessageQueueListener();

	private PeerTransferStats peerUploadStats_ = null;
	private PeerTransferStats peerDownloadStats_ = null;

	public XFerPeerListener(XFerLogPlugin plugin, Peer peer) {
		plugin_ = plugin;
		peer_ = peer;

		Connection conn = peer_.getConnection();
		if (conn != null) {
			IncomingMessageQueue qIn = conn.getIncomingMessageQueue();
			if (qIn != null)
				qIn.registerListener(incomingMsgListener_);
			OutgoingMessageQueue qOut = conn.getOutgoingMessageQueue();
			if (qOut != null)
				qOut.registerListener(outgoingMsgListener_);
		}
	}

	/* Handle connection state transitions */
	public void eventOccurred(PeerEvent event) {
		/* We only care about state changed events */
		if (event.getType() != PeerEvent.ET_STATE_CHANGED)
			return;

		/* Ensure the download still exists */
		Download download;
		try {
			download = peer_.getManager().getDownload();
		} catch (Exception e) {
			plugin_.getLogger().log("Failed to get reference to download", e);
			return;
		}

		Event logEvent = null;
		switch ((Integer)event.getData()) {
		case Peer.HANDSHAKING:
			logEvent = DownloadLogs.Event.EVT_PEER_HANDSHAKE;
			break;
		case Peer.TRANSFERING:
			logEvent = DownloadLogs.Event.EVT_PEER_CONNECTED;
			break;
		case Peer.CLOSING:
			logEvent = DownloadLogs.Event.EVT_PEER_CLOSING;
			break;
		}

		if (logEvent != null)
			plugin_.getLogProcessChain().push(
					PluginDownloadLogs.MakeLogRecordPeerEvent(plugin_, logEvent, download, peer_));
	}

	public void cleanup() {
		/* Remove the listeners */
		Connection conn = peer_.getConnection();
		if (conn != null) {
			IncomingMessageQueue qIn = conn.getIncomingMessageQueue();
			if (qIn != null)
				qIn.deregisterListener(incomingMsgListener_);
			OutgoingMessageQueue qOut = conn.getOutgoingMessageQueue();
			if (qOut != null)
				qOut.deregisterListener(outgoingMsgListener_);
		}

		/* Push out pending logs */
		if (peerUploadStats_ != null) {
			handleNewMessage(Event.EVT_XFER_UPLOAD_STATS, peerUploadStats_, true);
			peerUploadStats_ = null;
		}
		if (peerDownloadStats_ != null) {
			handleNewMessage(Event.EVT_XFER_DLOAD_STATS, peerDownloadStats_, true);
			peerDownloadStats_ = null;
		}
	}

}
