/**
 * Instrumentation Vuze Plugin
 * 
 * Copyright 2008, Yale Laboratory of Networked Systems
 */

package net.yale.cs.p4p.p2p.azureus.log.records;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.text.ParseException;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.zip.GZIPOutputStream;

import net.yale.cs.p4p.p2p.azureus.log.XFerLogPlugin;
import net.yale.cs.p4p.p2p.logs.DownloadLogs.LogRecord;

import org.gudy.azureus2.plugins.logging.LoggerChannel;

public class DownloadLogPoster extends PluginDownloadLogProcessor {

	private static final String INITIAL_PERSIST_FILE = "records1.dat";
	private static final String INITIAL_POSTING_FILE = "records2.dat";

	private Object needConfigLock_ = new Object();
	private RemoteLoggingCheck remoteLogCheck_;
	private boolean needConfig_ = true;

	/* Keep records that we currently know about on disk as well
	 * in case the client shuts down. */
	private Object persistLock_ = new Object();

	/* File stream currently used for persisting logs. New
	 * log records are automatically written to here. The
	 * records are also stored in memory in the queue that
	 * backs this DownloadLogPoster. */
	private File persistFile_ = null;
	private FileOutputStream persistStream_ = null;
	private List<LogRecord> persistRecords_ = new LinkedList<LogRecord>();

	/* File currently storing records that had been persisted
	 * but are currently being posted to the server. */
	private File postingFile_ = null;
	private FileOutputStream postingStream_ = null;
	private List<LogRecord> postingRecords_ = new LinkedList<LogRecord>();

	public DownloadLogPoster(XFerLogPlugin plugin, PluginDownloadLogProcessor next) {
		super(plugin, next, plugin.getConfig().getCfgLogPostInterval() * 1000);
		remoteLogCheck_ = new RemoteLoggingCheck(plugin);

		/* Initialize persistence files */
		persistFile_ = new File(plugin.getPluginInterface().getPluginDirectoryName(), INITIAL_PERSIST_FILE);
		postingFile_ = new File(plugin.getPluginInterface().getPluginDirectoryName(), INITIAL_POSTING_FILE);

		try {
			persistStream_ = new FileOutputStream(persistFile_);
			postingStream_ = new FileOutputStream(postingFile_);
		} catch (IOException e) {
			plugin.getLogger().log("Failed to open files for record persistence: " + e);
			disablePersist();
		}

		loadPersistRecords();
	}

	private void beforePost() {
		/* Swap the current buffers storing persisting records and posting
		 * records. */

		synchronized (persistLock_) {
			if (!isPersistEnabled())
				return;

			/* Warn if there are posting records that are being overwritten here. This
			 * should never happen. */
			if (!postingRecords_.isEmpty())
				getPlugin().getLogger().log("WARNING: overwriting records that have not yet been posted");
			
			/* Swap the files */
			File tempFile = persistFile_;
			persistFile_ = postingFile_;
			postingFile_ = tempFile;

			/* Swap the streams */
			FileOutputStream tempStream = persistStream_;
			persistStream_ = postingStream_;
			postingStream_ = tempStream;
			
			/* Swap the queues */
			List<LogRecord> tempRecords = persistRecords_;
			persistRecords_ = postingRecords_;
			postingRecords_ = tempRecords;
		}
	}

	private void afterPost() {
		/* Clear out the files that have just been posted. */
		
		synchronized (persistLock_) {
			
			if (!isPersistEnabled())
				return;
			
			try {
				clearFile(getPlugin().getLogger(), postingStream_, postingRecords_);
			} catch (IOException e) {
				getPlugin().getLogger().log("failed to clear posted records file; disabling logging", e);
				disablePersist();
			}
		}
	}

	private void disablePersist() {
		synchronized (persistLock_) {

			/* Close the streams */
			if (persistStream_ != null) {
				try {
					persistStream_.close();
				} catch (IOException e) { /* Ignore */ }
				persistStream_ = null;
				persistFile_ = null;
			}
			if (postingStream_ != null) {
				try {
					postingStream_.close();
				} catch (IOException e) { /* Ignore */ }
				postingStream_ = null;
				postingFile_ = null;
			}

			/* Erase the queues */
			if (persistRecords_ != null) {
				persistRecords_.clear();
				persistRecords_ = null;
			}
			if (postingRecords_ != null) {
				postingRecords_.clear();
				postingRecords_ = null;
			}
			
			getPlugin().getLogger().log("Finished disabling record persistence");
		}
	}

	private boolean isPersistEnabled() {
		synchronized (persistLock_) {
			return persistStream_ != null && postingStream_ != null;
		}
	}

	private static List<LogRecord> readFile(LoggerChannel logger, File file) throws IOException {
		List<LogRecord> result = new LinkedList<LogRecord>();

		BufferedReader input = new BufferedReader(new InputStreamReader(new FileInputStream(file)));
		try {
			String line;
			while ((line = input.readLine()) != null) {
				if ((line = line.trim()).length() == 0)
					continue;

				try {
					result.add(LogRecord.fromString(line));
				} catch (ParseException e) {
					logger.log("Parse exception while reading persisted record: " + e);
					continue;
				}
			}
		} finally {
			input.close();
		}
		
		return result;
	}

	private static void clearFile(LoggerChannel logger, FileOutputStream stream, List<LogRecord> records) throws IOException {
		/* Truncate the file */
		stream.getChannel().truncate(0);

		/* Clear out the store of records */
		records.clear();
	}

	private void loadPersistRecords() {
		/* Read records currently stored in persistence and posting files. If there are currently
		 * records stored in the posting file, load them into the persistence file. */
		synchronized (persistLock_) {
			
			if (!isPersistEnabled())
				return;
			
			try {
				/* Read existing persistence records into queue */
				persistRecords_.addAll(readFile(getPlugin().getLogger(), persistFile_));
				
				/* Read existing posting records and write them to the persistence file  */
				writePersistRecords(readFile(getPlugin().getLogger(), postingFile_));
				
				/* Clear current posting file */
				clearFile(getPlugin().getLogger(), postingStream_, postingRecords_);
				
			} catch (IOException e) {
				/* Failed to load, so something is wrong with the streams. Disable logging */
				disablePersist();
			}
		}
	}

	private void writePersistRecords(List<LogRecord> recs) {
		synchronized (persistLock_) {
			if (!isPersistEnabled())
				return;

			try {
				for (LogRecord rec : recs) {
					persistStream_.write((rec.toString() + "\n").getBytes());
					persistRecords_.add(rec);
				}
				persistStream_.flush();
			} catch (IOException e) {
				getPlugin().getLogger().log("failed to persist record to file: " + e);
			}
		}
	}

	private boolean getNeedConfig() {
		synchronized (needConfigLock_) {
			if (needConfig_) {
				/* Haven't yet sent the config; send it now
				 * and mark as no longer needed */
				needConfig_ = false;
				return true;
			} else {
				/* Already sent; don't need to send again */
				return false;
			}
		}
			
	}

	public void resetNeedConfig() {
		synchronized (needConfigLock_) {
			needConfig_ = true;
		}
	}

	private boolean isLogging() {
		/* Be sure to send the config again when logging is enabled again */

		if (!getPlugin().getConfig().getCfgEnabled()) {
			resetNeedConfig();
			return false;
		}

		if (!remoteLogCheck_.isEnabled()) {
			resetNeedConfig();
			return false;
		}

		return true;
	}

	
	public void push(List<LogRecord> logs) {
		/* NOTE: Ignore storing in backing queue since we keep our
		 * own queues in this object */
		
		if (!isLogging())
			return;
		
		if (getNeedConfig())
			writePersistRecords(Collections.singletonList(PluginDownloadLogs.MakeLogRecordConfigChanged(getPlugin())));

		writePersistRecords(logs);
	}

	public void push(LogRecord log) {
		push(Collections.singletonList(log));
	}

	protected void beforeProcess(boolean terminating) {
		/* Update logging status from remote server */
		remoteLogCheck_.doUpdateEnabled();
	}

	protected void process(boolean terminating, List<LogRecord> logs) {
		try {
			/* Prepare logs for posting */
			beforePost();

			/* Exit early if we aren't configured to send to the server */
			if (!isLogging())
				return;
			
			/* Exit early if persistence disabled */
			if (!isPersistEnabled())
				return;

			/* Extract records to be posted */
			List<LogRecord> postingRecords = new LinkedList<LogRecord>(postingRecords_);
			postingRecords_.clear();
			
			if (postingRecords.isEmpty())
				return;

			/* It's time to post the logs to the server */
			getPlugin().getLogger().log("posting log records to server");
			postRecords(terminating, getPlugin().getConfig().getCfgLogServerURL(), postingRecords);
		} catch (IOException e) {
			getPlugin().getLogger().log("failed to post log record to server: " + e);
			return;
		} finally {
			setInterval(getPlugin().getConfig().getCfgLogPostInterval() * 1000);
			afterPost();
		}
	}

	private static HttpURLConnection getConnection(URL url) throws IOException {
		URLConnection conn = url.openConnection();
		if (!(conn instanceof HttpURLConnection))
			throw new MalformedURLException("HTTP URL Required");
		return (HttpURLConnection)conn;
	}

	private void postRecords(boolean terminating, URL url, Collection<LogRecord> records) throws IOException {
		
		/* If configured, write the log records to the console */
		if (getPlugin().getConfig().getCfgLogPostedRecords() && getPlugin().getLogger().isEnabled())
			for (LogRecord rec : records)
				getPlugin().getLogger().log("Posting record: " + rec.toString());

		/* Open connection to the server */
		HttpURLConnection conn = getConnection(url);

		/* Hardcode smaller timeouts when terminating */
		int connectTimeout = getPlugin().getConfig().getCfgConnectTimeout();
		if (terminating)
			connectTimeout = Math.min(connectTimeout, 3);
		int readTimeout = getPlugin().getConfig().getCfgReadTimeout();
		if (terminating)
			readTimeout = Math.min(readTimeout, 3);
		conn.setConnectTimeout(connectTimeout * 1000);
		conn.setReadTimeout(readTimeout * 1000);

		/* Post to server */
		conn.setDoOutput(true);
		conn.setDoInput(true);
		conn.setRequestProperty("Content-Type", "text/plain");
		try {
			/* Write the log records, one per line */
			PrintWriter out = new PrintWriter(new GZIPOutputStream(conn.getOutputStream()));
			try {
				for (LogRecord rec : records) {
					out.write(rec.toString());
					out.write("\r\n");
				}
			} finally {
				out.close();
			}

			if (conn.getResponseCode() != HttpURLConnection.HTTP_OK)
				throw new IOException("Failed to post records to server; server returned status "
							+ conn.getResponseCode() + ": " + conn.getResponseMessage());
		} finally {
			conn.disconnect();
		}
	}
	
	public void stop() {
		disablePersist();
		
		super.stop();
	}

}
