/**
 * Common P2P Logging Classes
 * 
 * Copyright 2008, Yale Laboratory of Networked Systems
 */

package net.yale.cs.p4p.p2p.logs;

import java.text.ParseException;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.TreeMap;
import java.util.TreeSet;

import org.apache.commons.codec.DecoderException;
import org.apache.commons.codec.binary.Hex;

/* This class provides a threadsafe container for logging records. Logging records
 * are associated with a particular key. For example, all logging records
 * for a particular torrent are grouped together.
 * 
 * The data structures used to represent actual log records are inner classes.
 */
public class DownloadLogs {

	protected static final String nextTokenRequire(StringTokenizer st) throws ParseException {
		if (!st.hasMoreTokens())
			throw new ParseException("Insufficient number of tokens found", 0);
		return st.nextToken();
	}

	/* Enumeration of the events that we log */
	public static enum Event {
		/* Default event assigned at construction; it should
		 * be considered a bug for this to appear in any transmitted log */
		EVT_NONE,

		/* Transfer of bytes from a peer to self */
		EVT_XFER_DLOAD_STATS,

		/* Transfer of bytes from self to a peer */
		EVT_XFER_UPLOAD_STATS,

		/* Direct download statistics (i.e., how long it took) */
		EVT_DLOAD_STATS,

		/* Handshaking with another peer */
		EVT_PEER_HANDSHAKE,

		/* Connection established with another peer */
		EVT_PEER_CONNECTED,

		/* Closing connection with another peer */
		EVT_PEER_CLOSING,

		/* Configuration changed */
		EVT_CONFIG_CHANGED,
	}

	/* Enumeration of protocol type */
	public static enum Protocol {
		/* Default protocol assigned at construction; this is
		 * the default value if either TCP or UDP is not valid
		 * for a particular event type */
		PROT_NONE,
		
		/* TCP protocol */
		PROT_TCP,
		
		/* UDP protocol */
		PROT_UDP,
	}
	
	/* Enumeration of node type */
	public static enum NodeType {
		
		/* Default type assigned at construction, this is
		 * the default value if not valid for a particular
		 * event type. */
		NODE_NONE,
		
		/* Node is seeding */
		NODE_SEED,
		
		/* Node is a leeching */
		NODE_LEECH,
	}

	/* Encapsulate an algorithm's name and version as a single object. This
	 * allows the combination to be a key for a Map data structure. */
	public static class ComponentIdentity implements Comparable<ComponentIdentity> {
		private String name_;
		public final String getName() { return name_; }

		private String version_;
		public final String getVersion() { return version_; }

		public ComponentIdentity(String name, String version) {
			if (name == null || version == null)
				throw new IllegalArgumentException("name and version must not be non-null");
			name_ = name;
			version_ = version;
		}

		public boolean equals(Object o) {
			if (!(o instanceof ComponentIdentity))
				return false;

			ComponentIdentity oAlg = (ComponentIdentity)o;
			return name_.equals(oAlg.name_) && version_.equals(oAlg.version_);
		}

		public int hashCode() {
			return name_.hashCode() + version_.hashCode();
		}

		public int compareTo(ComponentIdentity o) {
			if (o == null)
				return 1;

			int rc = name_.compareTo(o.name_);
			if (rc != 0)
				return rc;
			
			return version_.compareTo(o.version_);
		}
	}

	/* Provide toString() and fromString() methods to encode/decode
	 * parameters as as string. Parameter names and values are comma-separated. */
	public static class ComponentParams {

		private Map<String, String> params_;
		public final Map<String, String> getParams() { return params_; }

		public ComponentParams(Map<String, String> params) {
			params_ = Collections.unmodifiableMap(params);
		}

		private ComponentParams() {
			params_ = new TreeMap<String, String>();
		}

		public static final ComponentParams fromString(String s) throws ParseException {
			ComponentParams result = new ComponentParams();
			if (s.length() == 0)
				return result;

			StringTokenizer st = new StringTokenizer(s, ",");
			while (st.hasMoreTokens()) {
				String key = nextTokenRequire(st);
				String value = nextTokenRequire(st);
				result.params_.put(key, value);
			}
			result.params_ = Collections.unmodifiableMap(result.params_);

			return result;
		}

		public final String toString() {
			StringBuffer sb = new StringBuffer();
			
			for (Map.Entry<String, String> e : getParams().entrySet())
				sb.append(',').append(e.getKey()).append(',').append(e.getValue());
			
			return sb.toString();
		}
	}

	/* Map of configured algorithms and their parameters. Provides toString() and fromString()
	 * methods to encode/decode as a single token. Algorithms are separated by semicolons, and
	 * name/version/parameters of a particular algorithm are separated by colons. */
	public static class Components extends TreeMap<ComponentIdentity, ComponentParams> {
		private static final long serialVersionUID = 1L;

		private static final ComponentParams EMPTY_PARAMS = new ComponentParams();

		public static final Components fromString(String s) throws ParseException {
			Components result = new Components();

			StringTokenizer st = new StringTokenizer(s, ";");

			/* Using size as the first token avoids a case where 
			 * the full token is the empty string. */
			Integer.parseInt(nextTokenRequire(st));

			/* Read name, version, and parameters for each algorithm */
			while (st.hasMoreTokens()) {
				StringTokenizer stComp = new StringTokenizer(st.nextToken(), ":");
				String name = nextTokenRequire(stComp);
				String version = nextTokenRequire(stComp);
				result.put(new ComponentIdentity(name, version),
						stComp.hasMoreTokens() ? ComponentParams.fromString(stComp.nextToken()) : EMPTY_PARAMS);
			}

			return result;
		}

		public final String toString() {
			StringBuffer sb = new StringBuffer(128);
		
			sb.append(size());

			for (Map.Entry<ComponentIdentity, ComponentParams> e : entrySet())
				sb.append(';')
					.append(e.getKey().getName())
					.append(':')
					.append(e.getKey().getVersion())
					.append(':')
					.append(e.getValue());

			return sb.toString();
		}

	}
	
	/* Collection of tags associated with a record */
	public static class Tags extends TreeSet<String> {
		private static final long serialVersionUID = 1L;

		public static final Tags fromString(String s) throws ParseException {
			Tags result = new Tags();

			StringTokenizer st = new StringTokenizer(s, ",");

			/* Using size as the first token avoids a case where 
			 * the full token is the empty string. */
			Integer.parseInt(nextTokenRequire(st));

			/* Read strings */
			while (st.hasMoreTokens())
				result.add(st.nextToken());

			return result;
		}

		public final String toString() {
			StringBuffer sb = new StringBuffer(128);
		
			sb.append(size());

			for (String t : this)
				sb.append(',').append(t);

			return sb.toString();
		}

	}

	/* Logging record. Provides functionality to convert to/from a line of raw text */
	public static class LogRecord {
		private static final Components EMPTY_COMPONENTS = new Components();
		private static final Tags EMPTY_TAGS = new Tags();
		private static final byte[] DEFAULT_HASH = new byte[] { 0 }; 
		private static final String DEFAULT_IP = "0.0.0.0";
		private static final String NULL_STRING = "null";

		private long receivedTime_ = 0;
		public final long getReceivedTime() { return receivedTime_; }
		public final void setReceivedTime(long value) { receivedTime_ = value; }

		private String logVersion_ = NULL_STRING;
		public final String getLogVersion() { return logVersion_; }
		public final void setLogVersion(String value) { logVersion_ = value != null ? value : NULL_STRING; }
		
		private Components components_ = EMPTY_COMPONENTS;
		public final Components getComponents() { return components_; }
		public final void setComponents(Components value) { components_ = value != null ? value : EMPTY_COMPONENTS; }

		private byte[] torrentId_ = DEFAULT_HASH;
		public final byte[] getTorrentId() { return torrentId_; }
		public final void setTorrentId(byte[] value) { torrentId_ = value != null ? value : DEFAULT_HASH; }

		private long fileSize_ = 0;
		public final long getFileSize() { return fileSize_; }
		public final void setFileSize(long value) { fileSize_ = value; }
		
		private long numLeechers_ = 0;
		public final long getNumLeechers() { return numLeechers_; }
		public final void setNumLeechers(long value) { numLeechers_ = value; }

		private long numSeeders_ = 0;
		public final long getNumSeeders() { return numSeeders_; }
		public final void setNumSeeders(long value) { numSeeders_ = value; }

		private byte[] selfId_ = DEFAULT_HASH;
		public final byte[] getSelfId() { return selfId_; }
		public final void setSelfId(byte[] value) { selfId_ = value != null ? value : DEFAULT_HASH; }

		private String selfIp_ = DEFAULT_IP;
		public final String getSelfIp() { return selfIp_; }
		public final void setSelfIp(String value) { selfIp_ = value != null ? value : DEFAULT_IP; }

		private int selfPort_ = 0;
		public final int getSelfPort() { return selfPort_; }
		public final void setSelfPort(int value) { selfPort_ = value; }
		
		private NodeType selfType_ = NodeType.NODE_NONE;
		public final NodeType getSelfType() { return selfType_; }
		public final void setSelfType(NodeType value) { selfType_ = value != null ? value : NodeType.NODE_NONE; }

		private Event event_ = Event.EVT_NONE;
		public final Event getEvent() { return event_; }
		public final void setEvent(Event value) { event_ = value != null ? value : Event.EVT_NONE; }

		private long startTime_ = 0;
		public final long getStartTime() { return startTime_; }
		public final void setStartTime(long value) { startTime_ = value; }

		private long endTime_ = 0;
		public final long getEndTime() { return endTime_; }
		public final void setEndTime(long value) { endTime_ = value; }

		private byte[] peerId_ = DEFAULT_HASH;
		public final byte[] getPeerId() { return peerId_; }
		public final void setPeerId(byte[] value) { peerId_ = value != null ? value : DEFAULT_HASH; }

		private String peerIp_ = DEFAULT_IP;
		public final String getPeerIp() { return peerIp_; }
		public final void setPeerIp(String value) { peerIp_ = value != null ? value : DEFAULT_IP; }

		private int peerPort_ = 0;
		public final int getPeerPort() { return peerPort_; }
		public final void setPeerPort(int value) { peerPort_ = value; }

		private NodeType peerType_ = NodeType.NODE_NONE;
		public final NodeType getPeerType() { return peerType_; }
		public final void setPeerType(NodeType value) { peerType_ = value != null ? value : NodeType.NODE_NONE; }

		private String peerKind_ = NULL_STRING;
		public final String getPeerKind() { return peerKind_; }
		public final void setPeerKind(String value) { peerKind_ = value != null ? value : NULL_STRING; }

		private Tags peerTags_ = EMPTY_TAGS;
		public final Tags getPeerTags() { return peerTags_; }
		public final void setPeerTags(Tags value) { peerTags_ = value; }

		private Protocol peerProtocol_ = Protocol.PROT_NONE;
		public final Protocol getPeerProtocol() { return peerProtocol_; }
		public final void setPeerProtocol(Protocol value) { peerProtocol_ = value; }

		private long peerUploadCap_ = 0;
		public final long getPeerUploadCap() { return peerUploadCap_; }
		public final void setPeerUploadCap(long value) { peerUploadCap_ = value; }

		private long peerDownloadCap_ = 0;
		public final long getPeerDownloadCap() { return peerDownloadCap_; }
		public final void setPeerDownloadCap(long value) { peerDownloadCap_ = value; }

		private long bytesRcvd_ = 0;
		public final long getBytesReceived() { return bytesRcvd_; }
		public final void setBytesReceived(long value) { bytesRcvd_ = value; }

		private long bytesSent_ = 0;
		public final long getBytesSent() { return bytesSent_; }
		public final void setBytesSent(long value) { bytesSent_ = value; }

		private long bytesDiscarded_ = 0;
		public final long getBytesDiscarded() { return bytesDiscarded_; }
		public final void setBytesDiscarded(long value) { bytesDiscarded_ = value; }

		private byte[] clientId_ = DEFAULT_HASH;
		public final byte[] getClientId() { return clientId_; }
		public final void setClientId(byte[] value) { clientId_ = value != null ? value : DEFAULT_HASH; }

		public static final LogRecord fromString(String s) throws ParseException {
			StringTokenizer st = new StringTokenizer(s);
			try {
				LogRecord rec = new LogRecord();
				rec.setReceivedTime(Long.parseLong(nextTokenRequire(st)));
				rec.setLogVersion(nextTokenRequire(st));
				rec.setComponents(Components.fromString(nextTokenRequire(st)));
				rec.setTorrentId(Hex.decodeHex(nextTokenRequire(st).toCharArray()));
				rec.setFileSize(Long.parseLong(nextTokenRequire(st)));
				rec.setNumLeechers(Long.valueOf(nextTokenRequire(st)));
				rec.setNumSeeders(Long.valueOf(nextTokenRequire(st)));
				rec.setSelfId(Hex.decodeHex(nextTokenRequire(st).toCharArray()));
				rec.setSelfIp(nextTokenRequire(st));
				rec.setSelfPort(Integer.parseInt(nextTokenRequire(st)));
				rec.setSelfType(NodeType.valueOf(nextTokenRequire(st)));
				rec.setEvent(Event.valueOf(nextTokenRequire(st)));
				rec.setStartTime(Long.parseLong(nextTokenRequire(st)));
				rec.setEndTime(Long.parseLong(nextTokenRequire(st)));
				rec.setPeerId(Hex.decodeHex(nextTokenRequire(st).toCharArray()));
				rec.setPeerIp(nextTokenRequire(st));
				rec.setPeerPort(Integer.parseInt(nextTokenRequire(st)));
				rec.setPeerType(NodeType.valueOf(nextTokenRequire(st)));
				rec.setPeerKind(nextTokenRequire(st));
				rec.setPeerTags(Tags.fromString(nextTokenRequire(st)));
				rec.setPeerProtocol(Protocol.valueOf(nextTokenRequire(st)));
				rec.setPeerUploadCap(Long.valueOf(nextTokenRequire(st)));
				rec.setPeerDownloadCap(Long.valueOf(nextTokenRequire(st)));
				rec.setBytesReceived(Long.parseLong(nextTokenRequire(st)));
				rec.setBytesSent(Long.parseLong(nextTokenRequire(st)));
				rec.setBytesDiscarded(Long.parseLong(nextTokenRequire(st)));

				/* New log records have the client ID */
				if (st.hasMoreTokens())
					rec.setClientId(Hex.decodeHex(nextTokenRequire(st).toCharArray()));
				return rec;
			} catch (NumberFormatException e) {
				throw new ParseException("Invalid number: " + e, 0);
			} catch (DecoderException e) {
				throw new ParseException("Invalid hex string: " + e, 0);
			}
		}
		
		public final String toString() {
			StringBuffer sb = new StringBuffer(128);
			sb.append(getReceivedTime())
				.append('\t')
				.append(getLogVersion())
				.append('\t')
				.append(getComponents())
				.append('\t')
				.append(Hex.encodeHex(getTorrentId()))
				.append('\t')
				.append(getFileSize())
				.append('\t')
				.append(getNumLeechers())
				.append('\t')
				.append(getNumSeeders())
				.append('\t')
				.append(Hex.encodeHex(getSelfId()))
				.append('\t')
				.append(getSelfIp())
				.append('\t')
				.append(getSelfPort())
				.append('\t')
				.append(getSelfType())
				.append('\t')
				.append(getEvent())
				.append('\t')
				.append(getStartTime())
				.append('\t')
				.append(getEndTime())
				.append('\t')
				.append(Hex.encodeHex(getPeerId()))
				.append('\t')
				.append(getPeerIp())
				.append('\t')
				.append(getPeerPort())
				.append('\t')
				.append(getPeerType())
				.append('\t')
				.append(getPeerKind())
				.append('\t')
				.append(getPeerTags())
				.append('\t')
				.append(getPeerProtocol())
				.append('\t')
				.append(getPeerUploadCap())
				.append('\t')
				.append(getPeerDownloadCap())
				.append('\t')
				.append(getBytesReceived())
				.append('\t')
				.append(getBytesSent())
				.append('\t')
				.append(getBytesDiscarded())
				.append('\t')
				.append(Hex.encodeHex(getClientId()));
			return sb.toString();
		}
	}

	private List<LogRecord> records_ = new LinkedList<LogRecord>();
	private Object recordsLock_ = new Object();

	public void addRecord(LogRecord rec) {
		addRecord(Collections.singletonList(rec));
	}

	public void addRecord(List<LogRecord> records) {
		synchronized (recordsLock_) {
			records_.addAll(records);
		}
	}

	public List<LogRecord> getRecords() {
		synchronized (recordsLock_) {
			List<LogRecord> prevRecords = records_;
			records_ = new LinkedList<LogRecord>();
			return prevRecords;
		}
	}

	public void exportTo(DownloadLogs target) {
		target.addRecord(getRecords());
	}

}
